# Copyright 2016 Ian Wienand (iwienand@redhat.com)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import os.path
import runpy
import sys

import diskimage_builder.paths


# borrowed from pip:locations.py
def running_under_virtualenv():
    """Return True if we're running inside a virtualenv, False otherwise."""
    if hasattr(sys, 'real_prefix'):
        return True
    elif sys.prefix != getattr(sys, "base_prefix", sys.prefix):
        return True
    return False


def activate_venv():
    if running_under_virtualenv():
        activate_this = os.path.join(sys.prefix, "bin", "activate_this.py")
        globs = runpy.run_path(activate_this, globals())
        globals().update(globs)
        del globs


def main():
    # If we are called directly from a venv install
    # (/path/venv/bin/disk-image-create) then nothing has added the
    # virtualenv bin/ dir to $PATH.  the exec'd script below will be
    # unable to find call other dib tools like dib-run-parts.
    #
    # One solution is to say that you should only ever run
    # disk-image-create in a shell that has already sourced
    # bin/activate.sh (all this really does is add /path/venv/bin to
    # $PATH).  That's not a great interface as resulting errors will
    # be very non-obvious.
    #
    # We can detect if we are running in a virtualenv and use
    # virtualenv's "activate_this.py" script to activate it ourselves
    # before we call the script.  This ensures we have the path setting
    activate_venv()

    environ = os.environ

    # pre-seed some paths for the shell script
    environ['_LIB'] = diskimage_builder.paths.get_path('lib')

    # export the path to the current python
    if not os.environ.get('DIB_PYTHON_EXEC'):
        os.environ['DIB_PYTHON_EXEC'] = sys.executable

    # we have to handle being called as "disk-image-create" or
    # "ramdisk-image-create".  ramdisk-iamge-create is just a symlink
    # to disk-image-create
    # XXX: we could simplify things by removing the symlink, and
    # just setting IS_RAMDISK in environ here depending on sys.argv[1]
    script = "%s/%s" % (diskimage_builder.paths.get_path('lib'),
                        os.path.basename(sys.argv[0]))

    os.execve("/bin/bash", ['bash', script] + sys.argv[1:], environ)



# Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import imp
import os
from oslotest import base

module_path = (os.path.dirname(os.path.realpath(__file__)) +
               '/../extra-data.d/10-merge-svc-map-files')
service_map = imp.load_source('service_map', module_path)


class TestDataMerge(base.BaseTestCase):

    nova_api_services = {
        u'nova-api': {
            u'debian': u'nova-api',
            u'default': u'openstack-nova-api'
        },
        u'nova-cert': {
            u'debian': u'nova-cert',
            u'default': u'openstack-nova-cert'
        },
        u'nova-compute': {
            u'debian': u'nova-compute',
            u'default': u'openstack-nova-compute'
        },
        u'nova-conductor': {
            u'debian': u'nova-conductor',
            u'default': u'openstack-nova-conductor'
        },
        u'nova-consoleauth': {
            u'debian': u'nova-console',
            u'default': u'openstack-nova-console'
        }
    }

    glance_api_services = {
        u'glance-api': {
            u'debian': u'glance-api',
            u'default': u'openstack-glance-api'
        },
        u'glance-reg': {
            u'debian': u'glance-reg',
            u'default': u'openstack-glance-registry'
        }
    }

    cinder_api_services = {
        u'cinder-api': {
            u'debian': u'cinder-api',
            u'default': u'openstack-cinder-api'
        },
        u'cinder-scheduler': {
            u'debian': u'cinder-scheduler',
            u'default': u'openstack-cinder-scheduler'
        }
    }

    def test_merge_data_fedora(self):

        fedora_nova_api_services = {
            u'nova-api': u'openstack-nova-api',
            u'nova-cert': u'openstack-nova-cert',
            u'nova-compute': u'openstack-nova-compute',
            u'nova-conductor': u'openstack-nova-conductor',
            u'nova-consoleauth': u'openstack-nova-console'
        }

        fedora_nova_glance_services = {
            u'nova-api': u'openstack-nova-api',
            u'nova-cert': u'openstack-nova-cert',
            u'nova-compute': u'openstack-nova-compute',
            u'nova-conductor': u'openstack-nova-conductor',
            u'nova-consoleauth': u'openstack-nova-console',
            u'glance-api': u'openstack-glance-api',
            u'glance-reg': u'openstack-glance-registry'
        }

        fedora_nova_glance_cinder_services = {
            u'nova-api': u'openstack-nova-api',
            u'nova-cert': u'openstack-nova-cert',
            u'nova-compute': u'openstack-nova-compute',
            u'nova-conductor': u'openstack-nova-conductor',
            u'nova-consoleauth': u'openstack-nova-console',
            u'glance-api': u'openstack-glance-api',
            u'glance-reg': u'openstack-glance-registry',
            u'cinder-api': u'openstack-cinder-api',
            u'cinder-scheduler': u'openstack-cinder-scheduler',
        }

        result = dict()
        result = service_map.merge_data(self.nova_api_services,
                                        result,
                                        "fedora")

        self.assertDictEqual(result,
                             fedora_nova_api_services,
                             "Merge failed")

        result = service_map.merge_data(self.glance_api_services,
                                        result,
                                        "fedora")

        self.assertDictEqual(result,
                             fedora_nova_glance_services,
                             "Merge failed")

        result = service_map.merge_data(self.cinder_api_services,
                                        result,
                                        "fedora")
        self.assertDictEqual(result,
                             fedora_nova_glance_cinder_services,
                             "Merge failed")

    def test_merge_data_ubuntu(self):

        ubuntu_nova_api_services = {
            u'nova-api': u'nova-api',
            u'nova-cert': u'nova-cert',
            u'nova-compute': u'nova-compute',
            u'nova-conductor': u'nova-conductor',
            u'nova-consoleauth': u'nova-console'
        }

        ubuntu_nova_glance_services = {
            u'nova-api': u'nova-api',
            u'nova-cert': u'nova-cert',
            u'nova-compute': u'nova-compute',
            u'nova-conductor': u'nova-conductor',
            u'nova-consoleauth': u'nova-console',
            u'glance-api': u'glance-api',
            u'glance-reg': u'glance-reg'
        }

        ubuntu_nova_glance_cinder_services = {
            u'nova-api': u'nova-api',
            u'nova-cert': u'nova-cert',
            u'nova-compute': u'nova-compute',
            u'nova-conductor': u'nova-conductor',
            u'nova-consoleauth': u'nova-console',
            u'glance-api': u'glance-api',
            u'glance-reg': u'glance-reg',
            u'cinder-api': u'cinder-api',
            u'cinder-scheduler': u'cinder-scheduler'
        }

        result = dict()
        result = service_map.merge_data(self.nova_api_services,
                                        result,
                                        "ubuntu")

        self.assertDictEqual(result,
                             ubuntu_nova_api_services,
                             "Merge failed")

        result = service_map.merge_data(self.glance_api_services,
                                        result,
                                        "ubuntu")

        self.assertDictEqual(result,
                             ubuntu_nova_glance_services,
                             "Merge failed")

        result = service_map.merge_data(self.cinder_api_services,
                                        result,
                                        "ubuntu")

        self.assertDictEqual(result,
                             ubuntu_nova_glance_cinder_services,
                             "Merge failed")



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import subprocess

from diskimage_builder.block_device.exception \
    import BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase
from diskimage_builder.block_device.utils import exec_sudo


logger = logging.getLogger(__name__)

#
# LVM
# ---
#
# The LVM config has three required keys; pvs, vgs and lvs
#
# lvm:     ->  LVSNode
#   pvs:   ->  PvsNode
#   lvs:   ->  LvsNode
#   vgs:   ->  VgsNode
#
# The LVMPlugin will verify this and build nodes into the
# configuration graph.
#
# As described below, a LVSNode is created for synchronisation
# purposes.  Thus if you had something like two partitions that became
# two physical-volumes (pv1 & pv2), that you then combine into a
# single volume group (vg) and then create several logical volumes
# (lv1, lv2, lv3) your graph would end up looking like:
#
#  partition1     partition2
#      |               |
#      ---> LVSNode <--+
#              |
#       +------+------+
#       v             v
#      pv1           pv2
#       |             |
#       +-->  vg  <---+
#             |
#       +-----+-----+
#       v     v     v
#      lv1   lv2   lv3
#
# After the create() call on the LVSNode object, the entire LVM setup
# would actually be complete.  The other nodes are all just
# place-holders, and are used for further ordering (for example, the
# fs creation & mounting should depend on the logical volume nodes).
# For this reason, their create() calls are blank.  However, for code
# organisational purposes they have a private _create() and _cleanup()
# call that is driven by the LVSNode object.


class PvsNode(NodeBase):
    def __init__(self, name, state, base, options):
        """Physical volume

        This is a placeholder node for the LVM physical volumes.

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: Parent partition
        :param options: config options
        """
        super(PvsNode, self).__init__(name, state)
        self.base = base
        self.options = options

    def _create(self):
        # the underlying device path of our parent was previously
        # recorded into the state during blockdev creation; look it
        # up.
        phys_dev = self.state['blockdev'][self.base]['device']

        cmd = ["pvcreate"]
        cmd.append(phys_dev)
        if self.options:
            cmd.extend(self.options)
        logger.debug("Creating pv command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        if 'pvs' not in self.state:
            self.state['pvs'] = {}
        self.state['pvs'][self.name] = {
            'opts': self.options,
            'device': phys_dev
        }

    def get_edges(self):
        # See LVMNode.get_edges() for how this gets connected
        return ([], [])

    def create(self):
        # see notes in LVMNode object
        pass


class VgsNode(NodeBase):
    def __init__(self, name, state, base, options):
        """Volume Group

        This is a placeholder node for a volume group

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: Parent :class:`PvsNodes` this volume group exists on
        :param options: extra options passed to the `vgcreate` command
        """
        super(VgsNode, self).__init__(name, state)
        self.base = base
        self.options = options

    def _create(self):
        # The PV's have saved their actual device name into the state
        # during their _create().  Look at our base elements and thus
        # find the underlying device paths in the state.
        pvs_devs = []
        for pv in self.base:
            pvs_dev = self.state['pvs'][pv]['device']
            pvs_devs.append(pvs_dev)

        cmd = ["vgcreate", ]
        cmd.append(self.name)
        cmd.extend(pvs_devs)
        if self.options:
            cmd.extend(self.options)

        logger.debug("Creating vg command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        if 'vgs' not in self.state:
            self.state['vgs'] = {}
        self.state['vgs'][self.name] = {
            'opts': self.options,
            'devices': self.base,
        }

    def _cleanup(self):
        exec_sudo(['vgchange', '-an', self.name])

    def get_edges(self):
        # self.base is already a list, per the config.  There might be
        # multiple pv parents here.
        edge_from = self.base
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        # see notes in LVMNode object
        pass


class LvsNode(NodeBase):
    def __init__(self, name, state, base, options, size, extents):
        """Logical Volume

        This is a placeholder node for a logical volume

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: the parent volume group
        :param options: options passed to lvcreate
        :param size: size of the LV, in MB (this or extents must be provided)
        :param extents: size of the LV in extents
        """
        super(LvsNode, self).__init__(name, state)
        self.base = base
        self.options = options
        self.size = size
        self.extents = extents

    def _create(self):
        cmd = ["lvcreate", ]
        cmd.extend(['--name', self.name])
        if self.size:
            cmd.extend(['-L', self.size])
        elif self.extents:
            cmd.extend(['-l', self.extents])
        if self.options:
            cmd.extend(self.options)

        cmd.append(self.base)

        logger.debug("Creating lv command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        self.state['blockdev'][self.name] = {
            'vgs': self.base,
            'size': self.size,
            'extents': self.extents,
            'opts': self.options,
            'device': '/dev/mapper/%s-%s' % (self.base, self.name)
        }

    def _cleanup(self):
        exec_sudo(['lvchange', '-an',
                   '/dev/%s/%s' % (self.base, self.name)])

    def get_edges(self):
        edge_from = [self.base]
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        # see notes in LVMNode object
        pass


class LVMNode(NodeBase):
    def __init__(self, name, state, pvs, lvs, vgs):
        """LVM Driver Node

        This is the "global" node where all LVM operations are driven
        from.  In the node graph, the LVM physical-volumes depend on
        this node.  This node then depends on the devices that the
        PV's require.  This node incorporates *all* LVM setup;
        i.e. after the create() call here we have created all pv's,
        lv's and vg.  The <Pvs|Lvs|Vgs>Node objects in the graph are
        therefore just dependency place holders whose create() call
        does nothing.

        Arguments:
        :param name: name of this node
        :param state: global state pointer
        :param pvs: A list of :class:`PvsNode` objects
        :param lvs: A list of :class:`LvsNode` objects
        :param vgs: A list of :class:`VgsNode` objects

        """
        super(LVMNode, self).__init__(name, state)
        self.pvs = pvs
        self.lvs = lvs
        self.vgs = vgs

    def get_edges(self):
        # This node requires the physical device(s), which is
        # recorded in the "base" argument of the PV nodes.
        pvs = []
        for pv in self.pvs:
            pvs.append(pv.base)
        edge_from = set(pvs)

        # The PV nodes should then depend on us.  i.e., we just made
        # this node a synchronisation point
        edge_to = [pv.name for pv in self.pvs]

        return (edge_from, edge_to)

    def create(self):
        # Run through pvs->vgs->lvs and create them
        # XXX: we could theoretically get this same info from walking
        # the graph of our children nodes?  Would that be helpful in
        # any way?
        for pvs in self.pvs:
            pvs._create()
        for vgs in self.vgs:
            vgs._create()
        for lvs in self.lvs:
            lvs._create()

    def cleanup(self):
        for lvs in self.lvs:
            lvs._cleanup()

        for vgs in self.vgs:
            vgs._cleanup()

        exec_sudo(['udevadm', 'settle'])


class LVMCleanupNode(NodeBase):
    def __init__(self, name, state, pvs):
        """Cleanup Node for LVM

        Information about the PV, VG and LV is typically
        cached in lvmetad. Even after removing PV device and
        partitions this data is not automatically updated
        which leads to a couple of problems.
        the 'pvscan --cache' scans the available disks
        and updates the cache.
        This must be called after the cleanup of the
        containing block device is done.
        """
        super(LVMCleanupNode, self).__init__(name, state)
        self.pvs = pvs

    def create(self):
        # This class is used for cleanup only
        pass

    def cleanup(self):
        try:
            exec_sudo(['pvscan', '--cache'])
        except subprocess.CalledProcessError as cpe:
            logger.debug("pvscan call result [%s]", cpe)

    def get_edges(self):
        # This node depends on all physical device(s), which is
        # recorded in the "base" argument of the PV nodes.
        edge_to = set()
        for pv in self.pvs:
            edge_to.add(pv.base)
        return ([], edge_to)


class LVMPlugin(PluginBase):

    def _config_error(self, msg):
        raise BlockDeviceSetupException(msg)

    def __init__(self, config, defaults, state):
        """Build LVM nodes

        This reads the "lvm:" config stanza, validates it and produces
        the PV, VG and LV nodes.  These are all synchronised via a
        LVMNode as described above.

        Arguments:
        :param config: "lvm" configuration dictionary
        :param defaults: global defaults dictionary
        :param state: global state reference
        """

        super(LVMPlugin, self).__init__()

        # note lvm: doesn't require a base ... the base is the
        # physical devices the "pvs" nodes are made on.
        if 'name' not in config:
            self._config_error("Lvm config requires 'name'")
        if 'pvs' not in config:
            self._config_error("Lvm config requires a 'pvs'")
        if 'vgs' not in config:
            self._config_error("Lvm config needs 'vgs'")
        if 'lvs' not in config:
            self._config_error("Lvm config needs 'lvs'")

        # create physical volume nodes
        self.pvs = []
        self.pvs_keys = []
        for pvs_cfg in config['pvs']:
            if 'name' not in pvs_cfg:
                self._config_error("Missing 'name' in pvs config")
            if 'base' not in pvs_cfg:
                self._config_error("Missing 'base' in pvs_config")

            pvs_item = PvsNode(pvs_cfg['name'], state,
                               pvs_cfg['base'],
                               pvs_cfg.get('options'))
            self.pvs.append(pvs_item)

        # create volume group nodes
        self.vgs = []
        self.vgs_keys = []
        for vgs_cfg in config['vgs']:
            if 'name' not in vgs_cfg:
                self._config_error("Missing 'name' in vgs config")
            if 'base' not in vgs_cfg:
                self._config_error("Missing 'base' in vgs config")

            # Ensure we have a valid PVs backing this VG
            for pvs in vgs_cfg['base']:
                if not any(pv.name == pvs for pv in self.pvs):
                    self._config_error("base:%s in vgs does not "
                                       "match a valid pvs" % pvs)
            vgs_item = VgsNode(vgs_cfg['name'], state, vgs_cfg['base'],
                               vgs_cfg.get('options', None))
            self.vgs.append(vgs_item)

        # create logical volume nodes
        self.lvs = []
        for lvs_cfg in config['lvs']:
            if 'name' not in lvs_cfg:
                self._config_error("Missing 'name' in lvs config")
            if 'base' not in lvs_cfg:
                self._config_error("Missing 'base' in lvs config")
            if 'size' not in lvs_cfg and 'extents' not in lvs_cfg:
                self._config_error("Missing 'size' or 'extents' in lvs config")

            # ensure this logical volume has a valid volume group base
            if not any(vg.name == lvs_cfg['base'] for vg in self.vgs):
                self._config_error("base:%s in lvs does not match a valid vg" %
                                   lvs_cfg['base'])

            lvs_item = LvsNode(lvs_cfg['name'], state, lvs_cfg['base'],
                               lvs_cfg.get('options', None),
                               lvs_cfg.get('size', None),
                               lvs_cfg.get('extents', None))
            self.lvs.append(lvs_item)

        # create the "driver" node
        self.lvm_node = LVMNode(config['name'], state,
                                self.pvs, self.lvs, self.vgs)
        self.lvm_cleanup_node = LVMCleanupNode(
            config['name'] + "-CLEANUP", state, self.pvs)

    def get_nodes(self):
        # the nodes for insertion into the graph are all of the pvs,
        # vgs and lvs nodes we have created above, the root node and
        # the cleanup node.
        return self.pvs + self.vgs + self.lvs \
            + [self.lvm_node, self.lvm_cleanup_node]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from diskimage_builder.block_device.exception import \
    BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase


logger = logging.getLogger(__name__)


class PartitionNode(NodeBase):

    flag_boot = 1
    flag_primary = 2

    def __init__(self, config, state, parent, prev_partition):

        super(PartitionNode, self).__init__(config['name'], state)

        self.base = config['base']
        self.partitioning = parent
        self.prev_partition = prev_partition

        self.flags = set()
        if 'flags' in config:
            for f in config['flags']:
                if f == 'boot':
                    self.flags.add(self.flag_boot)
                elif f == 'primary':
                    self.flags.add(self.flag_primary)
                else:
                    raise BlockDeviceSetupException("Unknown flag: %s" % f)

        if 'size' not in config:
            raise BlockDeviceSetupException("No size in partition" % self.name)
        self.size = config['size']

        self.ptype = int(config['type'], 16) if 'type' in config else 0x83

    def get_flags(self):
        return self.flags

    def get_size(self):
        return self.size

    def get_type(self):
        return self.ptype

    def get_edges(self):
        edge_from = [self.base]
        edge_to = []
        if self.prev_partition is not None:
            edge_from.append(self.prev_partition.name)
        return (edge_from, edge_to)

    # These all call back to the parent "partitioning" object to do
    # the real work.  Every node calls it, but only one will succeed;
    # see the gating we do in the parent function.
    #
    # XXX: A better model here would be for the parent object to a
    # real node in the config graph, so it's create() gets called.
    # These can then just be stubs.
    def create(self):
        self.partitioning.create()

    def cleanup(self):
        self.partitioning.cleanup()



# Copyright 2016 Ian Wienand (iwienand@redhat.com)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Export paths"""

import os
import pkg_resources
import sys


def get_path(var):
    if var == "lib":
        return os.path.abspath(
            pkg_resources.resource_filename(__name__, "lib"))
    elif var == "elements":
        return os.path.abspath(
            pkg_resources.resource_filename(__name__, "elements"))
    else:
        print("Unknown path request!")
        sys.exit(1)


def show_path(var):
    print(get_path(var))



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
import argparse
import collections
import errno
import logging
import os
import sys
import yaml

import diskimage_builder.logging_config

logger = logging.getLogger(__name__)


class MissingElementException(Exception):
    pass


class AlreadyProvidedException(Exception):
    pass


class MissingOSException(Exception):
    pass


class InvalidElementDir(Exception):
    pass


class Element(object):
    """An element"""

    def _get_element_set(self, path):
        """Get element set from element-[deps|provides] file

        Arguments:
        :param path: path to element description

        :return: the set of elements in the file, or a blank set if
                 the file is not found.
        """
        try:
            with open(path) as f:
                lines = (line.strip() for line in f)
                # Strip blanks, but do we want to strip comment lines
                # too?  No use case at the moment, and comments might
                # break other things that poke at the element-* files.
                lines = (line for line in lines if line)
                return set(lines)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return set([])
            else:
                raise

    def _make_rdeps(self, all_elements):
        """Make a list of reverse dependencies (who depends on us).

           Only valid after _find_all_elements()

           Arguments:
           :param all_elements: dict as returned by _find_all_elements()

           :return: nothing, but elements will have r_depends var
        """

        # note; deliberatly left out of __init__ so that accidental
        # access without init raises error
        self.r_depends = []
        for name, element in all_elements.items():
            if self.name in element.depends:
                self.r_depends.append(element.name)

    def __init__(self, name, path):
        """A new element

        :param name: The element name
        :param path: Full path to element.  element-deps and
                     element-provides files will be parsed
        """
        self.name = name
        self.path = path

        # read the provides & depends files for this element into a
        # set; if the element has them.
        self.provides = self._get_element_set(
            os.path.join(path, 'element-provides'))
        self.depends = self._get_element_set(
            os.path.join(path, 'element-deps'))

        logger.debug("New element : %s", str(self))

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return self.name

    def __str__(self):
        return '%s p:<%s> d:<%s>' % (self.name,
                                     ','.join(self.provides),
                                     ','.join(self.depends))


def _get_elements_dir():
    if not os.environ.get('ELEMENTS_PATH'):
        raise Exception("$ELEMENTS_PATH must be set.")
    return os.environ['ELEMENTS_PATH']


def _expand_element_dependencies(user_elements, all_elements):
    """Expand user requested elements using element-deps files.

    Arguments:
    :param user_elements: iterable enumerating the elements a user requested
    :param all_elements: Element object dictionary from find_all_elements

    :return: a set containing the names of user_elements and all
             dependent elements including any transitive dependencies.
    """
    final_elements = set(user_elements)
    check_queue = collections.deque(user_elements)
    provided = set()
    provided_by = collections.defaultdict(list)

    while check_queue:
        # bug #1303911 - run through the provided elements first to avoid
        # adding unwanted dependencies and looking for virtual elements
        element = check_queue.popleft()
        if element in provided:
            continue
        elif element not in all_elements:
            raise MissingElementException("Element '%s' not found" % element)

        element_obj = all_elements[element]

        element_deps = element_obj.depends
        element_provides = element_obj.provides
        # save which elements provide another element for potential
        # error message
        for provide in element_provides:
            provided_by[provide].append(element)
        provided.update(element_provides)
        check_queue.extend(element_deps - (final_elements | provided))
        final_elements.update(element_deps)

    conflicts = set(user_elements) & provided
    if conflicts:
        logger.error(
            "The following elements are already provided by another element")
        for element in conflicts:
            logger.error("%s : already provided by %s",
                         element, provided_by[element])
        raise AlreadyProvidedException()

    if "operating-system" not in provided:
        raise MissingOSException("Please include an operating system element")

    out = final_elements - provided
    return([all_elements[element] for element in out])


def _find_all_elements(paths=None):
    """Build a dictionary Element() objects

    Walk ELEMENTS_PATH and find all elements.  Make an Element object
    for each element we wish to consider.  Note we process overrides
    such that elements specified earlier in the ELEMENTS_PATH override
    those seen later.

    :param paths: A list of paths to find elements in.  If None will
                  use ELEMENTS_PATH from environment

    :return: a dictionary of all elements
    """

    all_elements = {}

    # note we process the later entries *first*, so that earlier
    # entries will override later ones.  i.e. with
    #  ELEMENTS_PATH=path1:path2:path3
    # we want the elements in "path1" to override "path3"
    if not paths:
        paths = list(reversed(_get_elements_dir().split(':')))
    else:
        paths = list(reversed(paths.split(':')))

    logger.debug("ELEMENTS_PATH is: %s", ":".join(paths))

    for path in paths:
        if not os.path.isdir(path):
            raise InvalidElementDir("ELEMENTS_PATH entry '%s' "
                                    "is not a directory " % path)

        # In words : make a list of directories in "path".  Since an
        # element is a directory, this is our list of elements.
        elements = [os.path.realpath(os.path.join(path, f))
                    for f in os.listdir(path)
                    if os.path.isdir(os.path.join(path, f))]

        for element in elements:
            # the element name is the last part of the full path in
            # element (these are all directories, we know that from
            # above)
            name = os.path.basename(element)

            new_element = Element(name, element)
            if name in all_elements:
                logger.warning("Element <%s> overrides <%s>",
                               new_element.path, all_elements[name].path)

            all_elements[name] = new_element

    # Now we have all the elements, make a call on each element to
    # store it's reverse dependencies
    for name, element in all_elements.items():
        element._make_rdeps(all_elements)

    return all_elements


def _get_elements(elements, paths=None):
    """Return the canonical list of Element objects

    This function returns Element objects.  For exernal calls, use
    get_elements which returns a simple tuple & list.

    :param elements: user specified list of elements
    :param paths: element paths, default to environment

    """
    all_elements = _find_all_elements(paths)
    return _expand_element_dependencies(elements, all_elements)


def get_elements(elements, paths=None):
    """Return the canonical list of elements with their dependencies

    .. note::

       You probably do not want to use this!  Elements that require
       access to the list of all other elements should generally use
       the environment variables exported by disk-image-create below.

    :param elements: user specified elements
    :param paths: Alternative ELEMENTS_PATH; default is to use from env

    :return: A de-duplicated list of tuples [(element, path),
             (element, path) ...] with all elements and their
             dependents, including any transitive dependencies.
    """

    elements = _get_elements(elements, paths)
    return [(element.name, element.path) for element in elements]


def expand_dependencies(user_elements, element_dirs):
    """Deprecated method for expanding element dependencies.

    .. warning::

       DO NOT USE THIS FUNCTION.  For compatibility reasons, this
       function does not provide paths to the returned elements.  This
       means the caller must process override rules if two elements
       with the same name appear in element_dirs

    :param user_elements: iterable enumerating the elements a user requested
    :param element_dirs: The ELEMENTS_PATH to process

    :return: a set containing user_elements and all dependent
             elements including any transitive dependencies.
    """
    logger.warning("expand_dependencies() deprecated, use get_elements")
    elements = _get_elements(user_elements, element_dirs)
    return set([element.name for element in elements])


def _output_env_vars(elements):
    """Output eval-able bash strings for IMAGE_ELEMENT vars

    :param elements: list of Element objects to represent
    """
    # first the "legacy" environment variable that just lists the
    # elements
    print("export IMAGE_ELEMENT='%s'" %
          ' '.join([element.name for element in elements]))

    # Then YAML
    output = {}
    for element in elements:
        output[element.name] = element.path
    print("export IMAGE_ELEMENT_YAML='%s'" % yaml.safe_dump(output))

    # Then bash array.  Unfortunately, bash can't export array
    # variables.  So we take a compromise and produce an exported
    # function that outputs the string to re-create the array.
    # You can then simply do
    #  eval declare -A element_array=$(get_image_element_array)
    # and you have it.
    output = ""
    for element in elements:
        output += '[%s]=%s ' % (element.name, element.path)
    print("function get_image_element_array {\n"
          "  echo \"%s\"\n"
          "};\n"
          "export -f get_image_element_array;" % output)


def main():
    diskimage_builder.logging_config.setup()

    parser = argparse.ArgumentParser()
    parser.add_argument('elements', nargs='+',
                        help='display dependencies of the given elements')
    parser.add_argument('--env', '-e', action='store_true',
                        default=False,
                        help=('Output eval-able bash strings for '
                              'IMAGE_ELEMENT variables'))

    args = parser.parse_args(sys.argv[1:])

    elements = _get_elements(args.elements)

    if args.env:
        _output_env_vars(elements)
    else:
        # deprecated compatibility output; doesn't include paths.
        print(' '.join([element.name for element in elements]))

    return 0

if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import logging
import os
import testtools
import yaml


logger = logging.getLogger(__name__)


class TestBase(testtools.TestCase):
    """Base for all test cases"""
    def setUp(self):
        super(TestBase, self).setUp()

        fs = '%(asctime)s %(levelname)s [%(name)s] %(message)s'
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.DEBUG, format=fs))

    def get_config_file(self, f):
        """Get the full path to sample config file f """
        logger.debug(os.path.dirname(__file__))
        return os.path.join(os.path.dirname(__file__), 'config', f)

    def load_config_file(self, f):
        """Load f and return it after yaml parsing"""
        path = self.get_config_file(f)
        with open(path, 'r') as config:
            return yaml.safe_load(config)



# Copyright 2016 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import random

from struct import pack


logger = logging.getLogger(__name__)


# Details of the MBR object itself can be found in the inline
# documentation.
#
# General design and implementation remarks:
# o Because the whole GNU parted and co. (e.g. the python-parted that
#   is based on GNU parted) cannot be used because of the license:
#   everything falls under GPL2 (not LGPL2!) and therefore does not
#   fit into the Apache License here.
# o It looks that there is no real alternative available (2016-06).
# o The interface of python-parted is not that simple to handle - and
#   the initial try to use GNU (python-)parted was not that much
#   easier and shorter than this approach.
# o When using tools (like fdisk or parted) they try to optimize the
#   alignment of partitions based on the data found on the host
#   system.  These might be misleading and might lead to (very) poor
#   performance.
# o These ready-to-use tools typically also change the CHS layout
#   based on the disk size.  In case that the disk is enlarged (which
#   is a normal use case for golden images), the CHS layout of the
#   disk changes for those tools (and is not longer correct).
#   In the DIB implementation the CHS are chosen that way, that also
#   for very small disks the maximum heads/cylinder and sectors/track
#   is used: even if the disk size in increased, the CHS numbers will
#   not change.
# o In the easy and straight forward way when only using one
#   partition, exactly 40 bytes (!) must be written - and the biggest
#   part of this data is fixed (same in all cases).
#
# Limitations and Incompatibilities
# o With the help of this class it is possible to create an
#   arbitrarily number of extended partitions (tested with over 1000).
# o There are limitations and shortcomings in the OS and in tools
#   handling these partitions.
# o Under Linux the loop device is able to handle a limited number of
#   partitions. The module parameter max_loop can be set - the maximum
#   number might vary depending on the distribution and kernel build.
# o Under Linux fdisk is able to handle 'only' 60 partitions. Only
#   those are listed, can be changed or written.
# o Under Linux GNU parted can handle about 60 partitions.
#
# Be sure only to pass in the number of partitions that the host OS
# and target OS are able to handle.

class MBR(object):
    """MBR Disk / Partition Table Layout

    Primary partitions are created first - and must also be passed in
    first.

    The extended partition layout is done in the way, that there is
    one entry in the MBR (the last) that uses the whole disk.
    EBR (extended boot records) are used to describe the partitions
    themselves.  This has the advantage, that the same procedure can
    be used for all partitions and arbitrarily many partitions can be
    created in the same way (the EBR is placed as block 0 in each
    partition itself).

    In conjunction with a fixed and 'fits all' partition alignment the
    major design focus is maximum performance for the installed image
    (vs. minimal size).

    Because of the chosen default alignment of 1MiB there will be
    (1MiB - 512B) unused disk space for the MBR and also the same
    size unused in every partition.

    Assuming that 512 byte blocks are used, the resulting layout for
    extended partitions looks like (blocks offset in extended
    partition given):

    ======== ==============================================
    Offset    Description
    ======== ==============================================
        0     MBR - 2047 blocks unused
     2048     EBR for partition 1 - 2047 blocks unused
     4096     Start of data for partition 1
     ...     ...
      X       EBR for partition N - 2047 blocks unused
      X+2048  Start of data for partition N
    ======== ==============================================

    Direct (native) writing of MBR, EBR (partition table) is
    implemented - no other partitioning library or tools is used -
    to be sure to get the correct CHS and alignment for a wide range
    of host systems.
    """

    # Design & Implementation details:
    # o A 'block' is a storage unit on disk. It is similar (equal) to a
    #   sector - but with LBA addressing.
    # o It is assumed that a disk block has that number of bytes
    bytes_per_sector = 512
    # o CHS is the 'good and very old way' specifying blocks.
    #   When passing around these numbers, they are also ordered like 'CHS':
    #   (cylinder, head, sector).
    # o The computation from LBA to CHS is not unique (it is based
    #   on the 'real' (or assumed) number of heads/cylinder and
    #   sectors/track), these are the assumed numbers.  Please note
    #   that these are also the maximum numbers:
    heads_per_cylinder = 254
    sectors_per_track = 63
    max_cylinders = 1023
    # o There is the need for some offsets that are defined in the
    #   MBR/EBR domain.
    MBR_offset_disk_id = 440
    MBR_offset_signature = 510
    MBR_offset_first_partition_table_entry = 446
    MBR_partition_type_extended_chs = 0x5
    MBR_partition_type_extended_lba = 0xF
    MBR_signature = 0xAA55

    def __init__(self, name, disk_size, alignment):
        """Initialize a disk partitioning MBR object.

        The name is the (existing) name of the disk.
        The disk_size is the (used) size of the disk. It must be a
        proper multiple of the disk bytes per sector (currently 512)
        """
        logger.info("Create MBR disk partitioning object")

        assert disk_size % MBR.bytes_per_sector == 0

        self.disk_size = disk_size
        self.disk_size_in_blocks \
            = self.disk_size // MBR.bytes_per_sector
        self.alignment_blocks = alignment // MBR.bytes_per_sector
        # Because the extended partitions are a chain of blocks, when
        # creating a new partition, the reference in the already
        # existing EBR must be updated. This holds a reference to the
        # latest EBR. (A special case is the first: when it points to
        # 0 (MBR) there is no need to update the reference.)
        self.disk_block_last_ref = 0

        self.name = name
        self.partition_abs_start = None
        self.partition_abs_next_free = None
        # Start of partition number
        self.partition_number = 0

        self.primary_partitions_created = 0
        self.extended_partitions_created = 0

    def __enter__(self):
        # Open existing file for writing (r+)
        self.image_fd = open(self.name, "r+b")
        self.write_mbr()
        self.write_mbr_signature(0)
        self.partition_abs_start = self.align(1)
        self.partition_abs_next_free \
            = self.partition_abs_start
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.image_fd.flush()
        os.fsync(self.image_fd.fileno())
        self.image_fd.close()

    def lba2chs(self, lba):
        """Converts a LBA block number to CHS

        If the LBA block number is bigger than the max (1023, 63, 254)
        the maximum is returned.
        """
        if lba > MBR.heads_per_cylinder * MBR.sectors_per_track \
           * MBR.max_cylinders:
            return MBR.max_cylinders, MBR.heads_per_cylinder, \
                MBR.sectors_per_track

        cylinder = lba // (MBR.heads_per_cylinder * MBR.sectors_per_track)
        head = (lba // MBR.sectors_per_track) % MBR.heads_per_cylinder
        sector = (lba % MBR.sectors_per_track) + 1

        logger.debug("Convert LBA to CHS [%d] -> [%d, %d, %d]",
                     lba, cylinder, head, sector)
        return cylinder, head, sector

    def encode_chs(self, cylinders, heads, sectors):
        """Encodes a CHS triple into disk format"""
        # Head - nothing to convert
        assert heads <= MBR.heads_per_cylinder
        eh = heads

        # Sector
        assert sectors <= MBR.sectors_per_track
        es = sectors
        # top two bits are set in cylinder conversion

        # Cylinder
        assert cylinders <= MBR.max_cylinders
        ec = cylinders % 256  # lower part
        hc = cylinders // 4   # extract top two bits and
        es = es | hc          # pass them into the top two bits of the sector

        logger.debug("Encode CHS to disk format [%d %d %d] "
                     "-> [%02x %02x %02x]", cylinders, heads, sectors,
                     eh, es, ec)
        return eh, es, ec

    def write_mbr(self):
        """Write MBR

        This method writes the MBR to disk. It creates a random disk
        id as well that it creates the extended partition (as
        first partition) which uses the whole disk.
        """
        disk_id = random.randint(0, 0xFFFFFFFF)
        self.image_fd.seek(MBR.MBR_offset_disk_id)
        self.image_fd.write(pack("<I", disk_id))

    def write_mbr_signature(self, blockno):
        """Writes the MBR/EBR signature to a block

        The signature consists of a 0xAA55 in the last two bytes of the
        block.
        """
        self.image_fd.seek(blockno *
                           MBR.bytes_per_sector +
                           MBR.MBR_offset_signature)
        self.image_fd.write(pack("<H", MBR.MBR_signature))

    def write_partition_entry(self, bootflag, blockno, entry, ptype,
                              lba_start, lba_length):
        """Writes a partition entry

        The entries are always the same and contain 16 bytes. The MBR
        and also the EBR use the same format.
        """
        logger.info("Write partition entry blockno [%d] entry [%d] "
                    "start [%d] length [%d]", blockno, entry,
                    lba_start, lba_length)

        self.image_fd.seek(
            blockno * MBR.bytes_per_sector +
            MBR.MBR_offset_first_partition_table_entry +
            16 * entry)
        # Boot flag
        self.image_fd.write(pack("<B", 0x80 if bootflag else 0x00))

        # Encode lba start / length into CHS
        chs_start = self.lba2chs(lba_start)
        chs_end = self.lba2chs(lba_start + lba_length)
        # Encode CHS into disk format
        chs_start_bin = self.encode_chs(*chs_start)
        chs_end_bin = self.encode_chs(*chs_end)

        # Write CHS start
        self.image_fd.write(pack("<BBB", *chs_start_bin))
        # Write partition type
        self.image_fd.write(pack("<B", ptype))
        # Write CHS end
        self.image_fd.write(pack("<BBB", *chs_end_bin))
        # Write LBA start & length
        self.image_fd.write(pack("<I", lba_start))
        self.image_fd.write(pack("<I", lba_length))

    def align(self, blockno):
        """Align the blockno to next alignment count"""
        if blockno % self.alignment_blocks == 0:
            # Already aligned
            return blockno

        return (blockno // self.alignment_blocks + 1) \
            * self.alignment_blocks

    def compute_partition_lbas(self, abs_start, size):
        lba_partition_abs_start = self.align(abs_start)
        lba_partition_rel_start \
            = lba_partition_abs_start - self.partition_abs_start
        lba_partition_length = size // MBR.bytes_per_sector
        lba_abs_partition_end \
            = self.align(lba_partition_abs_start + lba_partition_length)
        logger.info("Partition absolute [%d] relative [%d] "
                    "length [%d] absolute end [%d]",
                    lba_partition_abs_start, lba_partition_rel_start,
                    lba_partition_length, lba_abs_partition_end)
        return lba_partition_abs_start, lba_partition_length, \
            lba_abs_partition_end

    def add_primary_partition(self, bootflag, size, ptype):
        lba_partition_abs_start, lba_partition_length, lba_abs_partition_end \
            = self.compute_partition_lbas(self.partition_abs_next_free, size)

        self.write_partition_entry(
            bootflag, 0, self.partition_number, ptype,
            self.align(lba_partition_abs_start), lba_partition_length)

        self.partition_abs_next_free = lba_abs_partition_end
        logger.debug("Next free [%d]", self.partition_abs_next_free)
        self.primary_partitions_created += 1
        self.partition_number += 1
        return self.partition_number

    def add_extended_partition(self, bootflag, size, ptype):
        lba_ebr_abs = self.partition_abs_next_free
        logger.info("EBR block absolute [%d]", lba_ebr_abs)

        _, lba_partition_length, lba_abs_partition_end \
            = self.compute_partition_lbas(lba_ebr_abs + 1, size)

        # Write the reference to the new partition
        if self.disk_block_last_ref != 0:
            partition_complete_len = lba_abs_partition_end - lba_ebr_abs
            self.write_partition_entry(
                False, self.disk_block_last_ref, 1,
                MBR.MBR_partition_type_extended_chs,
                lba_ebr_abs - self.partition_abs_start,
                partition_complete_len)

        self.write_partition_entry(
            bootflag, lba_ebr_abs, 0, ptype, self.align(1),
            lba_partition_length)
        self.write_mbr_signature(lba_ebr_abs)

        self.partition_abs_next_free = lba_abs_partition_end
        logger.debug("Next free [%d]", self.partition_abs_next_free)
        self.disk_block_last_ref = lba_ebr_abs
        self.extended_partitions_created += 1
        self.partition_number += 1
        return self.partition_number

    def add_partition(self, primaryflag, bootflag, size, ptype):
        """Adds a partition with the given type and size"""
        logger.debug("Add new partition primary [%s] boot [%s] "
                     "size [%d] type [%x]",
                     primaryflag, bootflag, size, ptype)

        # primaries must be created before extended
        if primaryflag and self.extended_partitions_created > 0:
            raise RuntimeError("All primary partitions must be "
                               "given first")

        if primaryflag:
            return self.add_primary_partition(bootflag, size, ptype)
        if self.extended_partitions_created == 0:
            # When this is the first extended partition, the extended
            # partition entry has to be written.
            self.partition_abs_start = self.partition_abs_next_free
            self.write_partition_entry(
                False, 0, self.partition_number,
                MBR.MBR_partition_type_extended_lba,
                self.partition_abs_next_free,
                self.disk_size_in_blocks - self.partition_abs_next_free)
            self.partition_number = 4

        return self.add_extended_partition(bootflag, size, ptype)

    def free(self):
        """Returns the free (not yet partitioned) size"""
        return self.disk_size \
            - (self.partition_abs_next_free + self.align(1)) \
            * MBR.bytes_per_sector



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

import fixtures
import testtools

from diskimage_builder import element_dependencies

logger = logging.getLogger(__name__)

data_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'test-elements'))


def _populate_element(element_dir, element_name, element_deps=[], provides=[]):
    logger.debug("Populate %s <%s>", element_name, element_dir)
    element_home = os.path.join(element_dir, element_name)
    os.mkdir(element_home)
    deps_path = os.path.join(element_home, 'element-deps')

    with open(deps_path, 'w') as deps_file:
        deps_file.write("\n".join(element_deps))
        provides_path = os.path.join(element_home, 'element-provides')

    with open(provides_path, 'w') as provides_file:
        provides_file.write("\n".join(provides))


class TestElementDeps(testtools.TestCase):

    def setUp(self):
        super(TestElementDeps, self).setUp()
        self.element_root_dir = self.useFixture(fixtures.TempDir()).path

        self.element_dir = os.path.join(self.element_root_dir, 'elements')
        self.element_override_dir = os.path.join(self.element_root_dir,
                                                 'element-override')
        os.mkdir(self.element_dir)
        os.mkdir(self.element_override_dir)

        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.DEBUG))
        _populate_element(self.element_dir, 'requires-foo', ['foo'])
        _populate_element(self.element_dir,
                          'foo',
                          [],
                          ['operating-system'])
        _populate_element(self.element_dir,
                          'requires-requires-foo',
                          ['requires-foo'])
        _populate_element(self.element_dir, 'self', ['self'])
        _populate_element(self.element_dir,
                          'provides_virtual',
                          [],
                          ['virtual'])
        _populate_element(self.element_dir,
                          'requires_virtual',
                          ['virtual'],
                          ['operating-system'])
        _populate_element(self.element_dir, 'virtual', ['extra_dependency'])
        _populate_element(self.element_dir, 'extra_dependency', [])
        _populate_element(self.element_dir,
                          'circular1',
                          ['circular2'],
                          ['operating-system'])
        _populate_element(self.element_dir, 'circular2', ['circular1'])
        _populate_element(self.element_dir,
                          'provides_new_virtual',
                          [],
                          ['new_virtual', 'operating-system'])
        _populate_element(self.element_dir,
                          'requires_new_virtual',
                          ['new_virtual'])

        # second element should override the first one here
        _populate_element(self.element_dir, 'override_element', [])
        _populate_element(self.element_override_dir, 'override_element', [])

        # This simulates $ELEMENTS_PATH
        self.element_dirs = "%s:%s" % (self.element_override_dir,
                                       self.element_dir)

    # helper to return an (element, path) tuple from the standard dir
    def _e(self, element):
        return (element, os.path.join(self.element_dir, element))

    # helper to return an (element, path) tuple from the override dir
    def _eo(self, element):
        return (element, os.path.join(self.element_override_dir, element))

    def test_non_transitive_deps(self):
        result = element_dependencies.get_elements(['requires-foo'],
                                                   self.element_dirs)
        self.assertItemsEqual([self._e('foo'), self._e('requires-foo')],
                              result)

    def test_missing_deps(self):
        e = self.assertRaises(element_dependencies.MissingElementException,
                              element_dependencies.get_elements,
                              ['fake'],
                              self.element_dirs)
        self.assertIn("Element 'fake' not found", str(e))

    def test_invalid_element_dir(self):
        e = self.assertRaises(element_dependencies.InvalidElementDir,
                              element_dependencies.get_elements,
                              ['fake'],
                              self.element_dirs + ":/not/a/dir")
        self.assertIn("ELEMENTS_PATH entry '/not/a/dir' is not a directory",
                      str(e))

    def test_transitive_deps(self):
        result = element_dependencies.get_elements(
                ['requires-requires-foo'], self.element_dirs)

        self.assertItemsEqual([self._e('requires-requires-foo'),
                               self._e('requires-foo'),
                               self._e('foo')], result)

    def test_no_deps(self):
        result = element_dependencies.get_elements(['foo'], self.element_dirs)
        self.assertEqual([self._e('foo')], result)

    def test_self(self):
        result = element_dependencies.get_elements(['self', 'foo'],
                                                   self.element_dirs)
        self.assertItemsEqual([self._e('self'),
                               self._e('foo')], result)

    def test_circular(self):
        result = element_dependencies.get_elements(['circular1'],
                                                   self.element_dirs)
        self.assertItemsEqual([self._e('circular1'),
                               self._e('circular2')], result)

    def test_provide(self):
        result = element_dependencies.get_elements(
                ['provides_virtual', 'requires_virtual'],
                self.element_dirs)
        self.assertItemsEqual([self._e('requires_virtual'),
                               self._e('provides_virtual')], result)

    def test_provide_conflict(self):
        self.assertRaises(element_dependencies.AlreadyProvidedException,
                          element_dependencies.get_elements,
                                  ['virtual', 'provides_virtual'],
                                  self.element_dirs)

    def test_provide_virtual_ordering(self):
        result = element_dependencies.get_elements(
                ['requires_new_virtual', 'provides_new_virtual'],
                self.element_dirs)
        self.assertItemsEqual(
                [self._e('requires_new_virtual'),
                 self._e('provides_new_virtual')], result)

    def test_no_os_element(self):
        self.assertRaises(element_dependencies.MissingOSException,
                          element_dependencies.get_elements,
                          ['provides_virtual'],
                          self.element_dirs)

    def test_duplicated_os_passed_as_element(self):
        self.assertRaises(
                element_dependencies.AlreadyProvidedException,
                element_dependencies.get_elements,
                ['circular1', 'operating-system'],
                self.element_dirs)
        # ensure we get the error message about what's providing the
        # conflicting package
        self.assertIn("operating-system : already provided by ['circular1']",
                      self.log_fixture.output)

    def test_element_override(self):
        # make sure we picked up "override_element" from the override dir,
        # not the base dir
        result = element_dependencies.get_elements(['override_element', 'foo'],
                                                   self.element_dirs)
        self.assertItemsEqual([self._e('foo'),
                               self._eo('override_element')],
                              result)

    def test_expand_dependencies_deprecated(self):
        # test the deprecated expand_dependencies call
        result = element_dependencies.expand_dependencies(
                ['foo', 'requires-foo'], self.element_dirs)
        self.assertItemsEqual(['foo', 'requires-foo'], result)

    def test_output_sanity(self):
        # very basic output sanity test
        elements = element_dependencies._get_elements(['foo', 'requires-foo'],
                                                      self.element_dirs)
        element_dependencies._output_env_vars(elements)


class TestElements(testtools.TestCase):
    def test_depends_on_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable('ELEMENTS_PATH', '/foo/bar'))
        self.assertEqual('/foo/bar',
                         element_dependencies._get_elements_dir())

    def test_env_not_set(self):
        self.useFixture(fixtures.EnvironmentVariable('ELEMENTS_PATH', ''))
        self.assertRaises(Exception,
                          element_dependencies._get_elements_dir, ())



#!/usr/bin/env python
#
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is an output filter to filter and timestamp the logs from Grenade and
# DevStack. Largely our awk filters got beyond the complexity level which were
# sustainable, so this provides us much more control in a single place.
#
# The overhead of running python should be less than execing `date` a million
# times during a run.

import argparse
import datetime
import re
import sys

IGNORE_LINES = re.compile('(set \+o|xtrace)')


def get_options():
    parser = argparse.ArgumentParser(
        description='Filter output by DevStack and friends')
    parser.add_argument('-o', '--outfile',
                        help='Output file for content',
                        default=None)
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Write to stdout',
                        default=False)
    parser.add_argument('-b', '--no-timestamp', action='store_true',
                        help='Do not prefix stdout with timestamp (bare)',
                        default=False)
    return parser.parse_args()


def skip_line(line):
    """Should we skip this line."""
    return IGNORE_LINES.search(line) is not None


def main():
    opts = get_options()
    outfile = None
    if opts.outfile:
        outfile = open(opts.outfile, 'ab', 0)

    # Otherwise fileinput reprocess args as files
    sys.argv = []
    for line in iter(sys.stdin.readline, ''):
        # put skip lines here
        if skip_line(line):
            continue

        now = datetime.datetime.utcnow()
        ts_line = ("%s | %s" % (
            now.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
            line))

        if opts.verbose:
            sys.stdout.write(line if opts.no_timestamp else ts_line)
            sys.stdout.flush()
        if outfile:
            # We've opened outfile as a binary file to get the
            # non-buffered behaviour.  on python3, sys.stdin was
            # opened with the system encoding and made the line into
            # utf-8, so write the logfile out in utf-8 bytes.
            if sys.version_info < (3,):
                outfile.write(ts_line)
            else:
                outfile.write(ts_line.encode('utf-8'))
            outfile.flush()


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(1)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import networkx as nx
import os

from stevedore import extension

from diskimage_builder.block_device.exception import \
    BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase


logger = logging.getLogger(__name__)

_extensions = extension.ExtensionManager(
    namespace='diskimage_builder.block_device.plugin',
    invoke_on_load=False)


# check if a given name is registered as a plugin
def is_a_plugin(name):
    return any(
        _extensions.map(lambda x: x.name == name))


def recurse_config(config, parent_base=None):
    """Convert a config "tree" to it's canonical name/base graph version

    This is a recursive function to convert a YAML layout "tree"
    config into a "flat" graph-based config.

    Arguments:
    :param config: the incoming config dictionary
    :param parent_base: the name of the parent node, if any
    :return: a list of expanded, graph-based config items

    """
    output = []
    this = {}

    # We should only have one key, with multiple values, being the
    # config entries.  e.g. (this was checked by config_tree_to_graph)
    #  mkfs:
    #    type: ext4
    #    label: 1234
    assert len(config.items()) == 1
    for k, v in config.items():
        key = k
        values = v

    # If we don't have a base, we take the parent base; first element
    # can have no base, however.
    if 'base' not in values:
        if parent_base is not None:
            this['base'] = parent_base
    else:
        this['base'] = values['base']

    # If we don't have a name, it is made up as "key_base"
    if 'name' not in values:
        this['name'] = "%s_%s" % (key, this['base'])
    else:
        this['name'] = values['name']

    # Go through the the values dictionary.  Either this is a "plugin"
    # key that needs to be recursed, or it is a value that is part of
    # this config entry.
    for nk, nv in values.items():
        if nk == "partitions":
            # "partitions" is a special key of the "partitioning"
            # object.  It is a list.  Each list-entry gets treated
            # as a top-level entry, so we need to recurse it's
            # keys.  But instead of becoming its own entry in the
            # graph, it gets attached to the .partitions attribute
            # of the parent. (see end for example)
            this['partitions'] = []
            for partition in nv:
                new_part = {}
                for pk, pv in partition.items():
                    if is_a_plugin(pk):
                        output.extend(
                            recurse_config({pk: pv}, partition['name']))
                    else:
                        new_part[pk] = pv
                new_part['base'] = this['base']
                this['partitions'].append(new_part)
        elif is_a_plugin(nk):
            # is this key a plugin directive?  If so, we recurse
            # into it.
            output.extend(recurse_config({nk: nv}, this['name']))
        else:
            # A value entry; just save as part of this entry
            this[nk] = nv

    output.append({k: this})
    return output


def config_tree_to_graph(config):
    """Turn a YAML config into a graph config

    Our YAML config is a list of entries.  Each

    Arguments:
    :parm config: YAML config; either graph or tree
    :return: graph-based result

    """
    output = []

    for entry in config:
        # Top-level entries should be a dictionary and have a plugin
        # registered for it
        if not isinstance(entry, dict):
            raise BlockDeviceSetupException(
                "Config entry not a dict: %s" % entry)

        keys = list(entry.keys())

        if len(keys) != 1:
            raise BlockDeviceSetupException(
                "Config entry top-level should be a single dict: %s" % entry)

        if not is_a_plugin(keys[0]):
            raise BlockDeviceSetupException(
                "Config entry is not a plugin value: %s" % entry)

        output.extend(recurse_config(entry))

    return output


def create_graph(config, default_config, state):
    """Generate configuration digraph

    Generate the configuration digraph from the config

    :param config: graph configuration file
    :param default_config: default parameters (from --params)
    :param state: reference to global state dictionary.
      Passed to :func:`PluginBase.__init__`
    :return: tuple with the graph object (a :class:`nx.Digraph`),
      ordered list of :class:`NodeBase` objects

    """
    # This is the directed graph of nodes: each parse method must
    # add the appropriate nodes and edges.
    dg = nx.DiGraph()

    for config_entry in config:
        # this should have been checked by generate_config
        assert len(config_entry) == 1

        logger.debug("Config entry [%s]", config_entry)
        cfg_obj_name = list(config_entry.keys())[0]
        cfg_obj_val = config_entry[cfg_obj_name]

        # Instantiate a "plugin" object, passing it the
        # configuration entry
        # XXX : would a "factory" pattern for plugins, where we
        # make a method call on an object stevedore has instantiated
        # be better here?
        if not is_a_plugin(cfg_obj_name):
            raise BlockDeviceSetupException(
                ("Config element [%s] is not implemented" % cfg_obj_name))
        plugin = _extensions[cfg_obj_name].plugin
        assert issubclass(plugin, PluginBase)
        cfg_obj = plugin(cfg_obj_val, default_config, state)

        # Ask the plugin for the nodes it would like to insert
        # into the graph.  Some plugins, such as partitioning,
        # return multiple nodes from one config entry.
        nodes = cfg_obj.get_nodes()
        assert isinstance(nodes, list)
        for node in nodes:
            # plugins should return nodes...
            assert isinstance(node, NodeBase)
            # ensure node names are unique.  networkx by default
            # just appends the attribute to the node dict for
            # existing nodes, which is not what we want.
            if node.name in dg.node:
                raise BlockDeviceSetupException(
                    "Duplicate node name: %s" % (node.name))
            logger.debug("Adding %s : %s", node.name, node)
            dg.add_node(node.name, obj=node)

    # Now find edges
    for name, attr in dg.nodes(data=True):
        obj = attr['obj']
        # Unfortunately, we can not determine node edges just from
        # the configuration file.  It's not always simply the
        # "base:" pointer.  So ask nodes for a list of nodes they
        # want to point to.  *mostly* it's just base: ... but
        # mounting is different.
        #  edges_from are the nodes that point to us
        #  edges_to are the nodes we point to
        edges_from, edges_to = obj.get_edges()
        logger.debug("Edges for %s: f:%s t:%s", name,
                     edges_from, edges_to)
        for edge_from in edges_from:
            if edge_from not in dg.node:
                raise BlockDeviceSetupException(
                    "Edge not defined: %s->%s" % (edge_from, name))
            dg.add_edge(edge_from, name)
        for edge_to in edges_to:
            if edge_to not in dg.node:
                raise BlockDeviceSetupException(
                    "Edge not defined: %s->%s" % (name, edge_to))
            dg.add_edge(name, edge_to)

    # this can be quite helpful debugging but needs pydotplus which
    # isn't in requirements.  for debugging, do
    #   .tox/py27/bin/pip install pydotplus
    #   DUMP_CONFIG_GRAPH=1 tox -e py27 -- specific_test
    #   dotty /tmp/graph_dump.dot
    # to see helpful output
    if 'DUMP_CONFIG_GRAPH' in os.environ:
        nx.nx_pydot.write_dot(dg, '/tmp/graph_dump.dot')

    # Topological sort (i.e. create a linear array that satisfies
    # dependencies) and return the object list
    call_order_nodes = nx.topological_sort(dg)
    logger.debug("Call order: %s", list(call_order_nodes))
    call_order = [dg.node[n]['obj'] for n in call_order_nodes]

    return dg, call_order

#
# On partitioning: objects
#
# To be concrete --
#
# partitioning:
#   base: loop0
#   name: mbr
#   partitions:
#     - name: partition1
#       foo: bar
#       mkfs:
#         type: xfs
#         mount:
#           mount_point: /
#
# gets turned into the following graph:
#
# partitioning:
#   partitions:
#     - name: partition1
#       base: image0
#       foo: bar
#
# mkfs:
#   base: partition1
#   name: mkfs_partition1
#   type: xfs
#
# mount:
#   base: mkfs_partition1
#   name: mount_mkfs_partition1
#   mount_point: /



# Copyright 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import time

from diskimage_builder.tests import base


class TestCacheUrl(base.ScriptTestBase):
    def test_cache_url_caches(self):
        tempdir = tempfile.mkdtemp()
        target = os.path.join(tempdir, 'target')
        source = 'http://fake/url'
        # Write fake data to the target file and return success
        self._stub_script('curl', 'echo "test" > ${3:7:100}\necho 200')
        self._run_command(['elements/cache-url/bin/cache-url',
                           source,
                           target])
        self.assertTrue(os.path.exists(target))
        modification_time = os.path.getmtime(target)
        # Make sure that the timestamp would change if the file does
        time.sleep(1)
        self._stub_script('curl', 'echo "304"')
        self._run_command(['elements/cache-url/bin/cache-url',
                           source,
                           target])
        self.assertEqual(modification_time, os.path.getmtime(target))



# Copyright 2016 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Python Logging Configuration for DIB
# Usage:
# In the main (application) file, do an
#   import logging_config
#   ...
#   logging_config.setup()
# It is then possible to use the normal python logging interface, like
#   logger = logging.getLogger(__name__)
#   logger.info("Info Message")

import json
import logging
import logging.config
import os
import sys


# A simple formatter to more or less copy oslo.log's ContextFormatter
class DibFormatter(logging.Formatter):

    def __init__(self, *args, **kwargs):
        self.fmt = kwargs['fmt']
        self.debug_suffix = kwargs.pop('debug_suffix')
        logging.Formatter.__init__(self, *args, **kwargs)

    def format(self, record):
        if record.levelno == logging.DEBUG:
            fmt = self.fmt + self.debug_suffix
        else:
            fmt = self.fmt

        if sys.version_info < (3, 2):
            self._fmt = fmt
        else:
            self._style = logging.PercentStyle(fmt)
            self._fmt = self._style._fmt

        return logging.Formatter.format(self, record)


def setup():
    # Check for the DIB_DEBUG_TRACE environment variable
    # If it is set to something greater then 0, use debug
    # logging.
    LOG_LEVEL = "DEBUG" if 'DIB_DEBUG_TRACE' in os.environ \
                and int(os.environ['DIB_DEBUG_TRACE']) > 0 else "INFO"

    # Default logging configuration which can be overwritten
    # by a config file passed in by a user.
    PYTHON_LOGGING_CONFIG = {
        'version': 1,
        'disable_existing_loggers': False,

        # note that disk-image-create runs stdout through
        # outfilter.py, which adds the timestamp.  this doesn't have a
        # timestamp to avoid double logging
        'formatters': {
            'standard': {
                '()': 'diskimage_builder.logging_config.DibFormatter',
                'fmt': '%(levelname)s '
                '%(name)s [-] %(message)s',
                'debug_suffix': ' %(funcName)s %(pathname)s:%(lineno)d'
            }
        },
        'handlers': {
            'default': {
                'level': LOG_LEVEL,
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
            }
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': LOG_LEVEL,
                'propagate': True
            }
        }
    }

    if 'DIB_PYTHON_LOGGING_CONFIG_FILE' in os.environ:
        with open(os.environ['DIB_PYTHON_LOGGING_CONFIG_FILE']) as fd:
            PYTHON_LOGGING_CONFIG = json.load(fd)

    logging.config.dictConfig(PYTHON_LOGGING_CONFIG)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import logging
import mock
import os
import subprocess

import diskimage_builder.block_device.tests.test_base as tb

from diskimage_builder.block_device.level0.localloop import image_create
from diskimage_builder.block_device.level1.mbr import MBR


logger = logging.getLogger(__name__)


class TestMBR(tb.TestBase):

    disk_size_10M = 10 * 1024 * 1024
    disk_size_1G = 1024 * 1024 * 1024

    def _get_path_for_partx(self):
        """Searches and sets the path for partx

        Because different distributions store the partx binary
        at different places, there is the need to look for it.
        """

        dirs = ["/bin", "/usr/bin", "/sbin", "/usr/sbin"]

        for d in dirs:
            if os.path.exists(os.path.join(d, "partx")):
                return os.path.join(d, "partx")
                return
        # If not found, try without path.
        return "partx"

    def setUp(self):
        super(TestMBR, self).setUp()

        self.tmp_dir = fixtures.TempDir()
        self.useFixture(self.tmp_dir)
        self.image_path = os.path.join(self.tmp_dir.path, "image.raw")
        image_create(self.image_path, TestMBR.disk_size_1G)
        logger.debug("Temp image is %s", self.image_path)

        self.partx_args = [self._get_path_for_partx(), "--raw",
                           "--output", "NR,START,END,TYPE,FLAGS,SCHEME",
                           "-g", "-b", "-", self.image_path]

    def _run_partx(self, image_path):
        logger.info("Running command: %s", self.partx_args)
        return subprocess.check_output(self.partx_args).decode("ascii")

    @mock.patch('os.fsync', wraps=os.fsync)
    def test_one_ext_partition(self, mock_os_fsync):
        """Creates one partition and check correctness with partx."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            mbr.add_partition(False, False, TestMBR.disk_size_10M, 0x83)

        # the exit handler of MBR should have synced the raw device
        # before exit
        mock_os_fsync.assert_called()

        output = self._run_partx(self.image_path)
        self.assertEqual(
            "1 2048 2097151 0xf 0x0 dos\n"
            "5 4096 24575 0x83 0x0 dos\n", output)

    def test_zero_partitions(self):
        """Creates no partition and check correctness with partx."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024):
            pass
        output = self._run_partx(self.image_path)
        self.assertEqual("", output)

    def test_many_ext_partitions(self):
        """Creates many partition and check correctness with partx."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            for nr in range(0, 64):
                mbr.add_partition(False, False, TestMBR.disk_size_10M, 0x83)

        output = self._run_partx(self.image_path)

        lines = output.split("\n")
        self.assertEqual(66, len(lines))

        self.assertEqual(
            "1 2048 2097151 0xf 0x0 dos", lines[0])

        start_block = 4096
        end_block = start_block + TestMBR.disk_size_10M / 512 - 1
        for nr in range(1, 65):
            fields = lines[nr].split(" ")
            self.assertEqual(6, len(fields))
            self.assertEqual(nr + 4, int(fields[0]))
            self.assertEqual(start_block, int(fields[1]))
            self.assertEqual(end_block, int(fields[2]))
            self.assertEqual("0x83", fields[3])
            self.assertEqual("0x0", fields[4])
            self.assertEqual("dos", fields[5])
            start_block += 22528
            end_block = start_block + TestMBR.disk_size_10M / 512 - 1

    def test_one_pri_partition(self):
        """Creates one primary partition and check correctness with partx."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            mbr.add_partition(True, False, TestMBR.disk_size_10M, 0x83)

        output = self._run_partx(self.image_path)
        self.assertEqual(
            "1 2048 22527 0x83 0x0 dos\n", output)

    def test_three_pri_partition(self):
        """Creates three primary partition and check correctness with partx."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            for _ in range(3):
                mbr.add_partition(True, False, TestMBR.disk_size_10M, 0x83)

        output = self._run_partx(self.image_path)
        self.assertEqual(
            "1 2048 22527 0x83 0x0 dos\n"
            "2 22528 43007 0x83 0x0 dos\n"
            "3 43008 63487 0x83 0x0 dos\n", output)

    def test_many_pri_and_ext_partition(self):
        """Creates many primary and extended partitions."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            # Create three primary partitions
            for _ in range(3):
                mbr.add_partition(True, False, TestMBR.disk_size_10M, 0x83)
            for _ in range(7):
                mbr.add_partition(False, False, TestMBR.disk_size_10M, 0x83)

        output = self._run_partx(self.image_path)
        self.assertEqual(
            "1 2048 22527 0x83 0x0 dos\n"     # Primary 1
            "2 22528 43007 0x83 0x0 dos\n"    # Primary 2
            "3 43008 63487 0x83 0x0 dos\n"    # Primary 3
            "4 63488 2097151 0xf 0x0 dos\n"   # Extended
            "5 65536 86015 0x83 0x0 dos\n"    # Extended Partition 1
            "6 88064 108543 0x83 0x0 dos\n"   # Extended Partition 2
            "7 110592 131071 0x83 0x0 dos\n"  # ...
            "8 133120 153599 0x83 0x0 dos\n"
            "9 155648 176127 0x83 0x0 dos\n"
            "10 178176 198655 0x83 0x0 dos\n"
            "11 200704 221183 0x83 0x0 dos\n", output)

    def test_pri_fat32_lba_partition(self):
        """Creates a partition with a non-default 'type' and verifies."""

        with MBR(self.image_path, TestMBR.disk_size_1G, 1024 * 1024) as mbr:
            mbr.add_partition(True, False, TestMBR.disk_size_10M, 0x0c)

        output = self._run_partx(self.image_path)
        self.assertEqual(
            "1 2048 22527 0xc 0x0 dos\n", output)
